"""Tests for interface class."""

from aiohttp import ClientSession, CookieJar
from aiohttp.client_exceptions import InvalidUrlClientError
from aiohttp.test_utils import TestClient, TestServer, loop_context
from unittest.mock import patch
import pytest
from typing import Any

from aioptdevices.configuration import Configuration
from aioptdevices.errors import (
    PTDevicesForbiddenError,
    PTDevicesRequestError,
    PTDevicesUnauthorizedError,
)
from aioptdevices.interface import Interface, PTDevicesResponse

# Add your token and device ID, **keep out of git**
from .mock_api import (
    API_URL,
    BAD_GATEWAY_ERROR_DEVICE_ID,
    EMPTY_ERROR_DEVICE_ID,
    FORBIDDEN_ERROR_DEVICE_ID,
    GOOD_RESP_DICT,
    GOOD_RESP_DICT_US_IMP,
    GOOD_RESP_DICT_UK_IMP,
    NORMAL_DEVICE_ID,
    NORMAL_DEVICE_ID_US,
    NORMAL_DEVICE_ID_UK,
    NORMAL_DEVICE_MAC,
    REDIRECT_ERROR_DEVICE_ID,
    TOKEN,
    UNAUTHORIZED_ERROR_DEVICE_ID,
    WRONG_CONTENT_TYPE_DEVICE_ID,
)
from .secret import DEVICE_ID as secret_DEVICE_ID, TOKEN as secret_TOKEN

# Query Paramaters


def test_interface(test_web_app):
    """Test the basic interface connection."""
    with loop_context() as loop:
        # Setup the test server
        server = TestServer(test_web_app)
        client = TestClient(server, loop=loop)
        loop.run_until_complete(client.start_server())

        normal_config: Configuration = Configuration(
            auth_token=TOKEN,
            device_id=NORMAL_DEVICE_ID,
            url=str(client.make_url(API_URL)),
            session=client.session,
        )
        interface: Interface = Interface(normal_config)

        async def test_get_data():
            resp: PTDevicesResponse = await interface.get_data()
            assert resp.get("body") == GOOD_RESP_DICT

        loop.run_until_complete(test_get_data())
        loop.run_until_complete(client.close())


def test_mac_id(test_web_app):
    """Test the basic interface connection."""
    with loop_context() as loop:
        # Setup the test server
        server = TestServer(test_web_app)
        client = TestClient(server, loop=loop)
        loop.run_until_complete(client.start_server())

        normal_config: Configuration = Configuration(
            auth_token=TOKEN,
            device_id=NORMAL_DEVICE_MAC,
            url=str(client.make_url(API_URL)),
            session=client.session,
        )
        interface: Interface = Interface(normal_config)

        async def test_get_data():
            resp: PTDevicesResponse = await interface.get_data()
            assert resp.get("body") == GOOD_RESP_DICT

        loop.run_until_complete(test_get_data())
        loop.run_until_complete(client.close())


def test_unit_conversion(test_web_app):
    """Test the automatic unit conversion."""
    with loop_context() as loop:
        # Setup the test server
        server = TestServer(test_web_app)
        client = TestClient(server, loop=loop)
        loop.run_until_complete(client.start_server())

        # Test US Units
        normal_config: Configuration = Configuration(
            auth_token=TOKEN,
            device_id=NORMAL_DEVICE_ID_US,
            url=str(client.make_url(API_URL)),
            session=client.session,
        )
        interface: Interface = Interface(normal_config)

        async def test_get_data_us():
            resp: PTDevicesResponse = await interface.get_data()
            assert resp.get("body") == GOOD_RESP_DICT_US_IMP

        loop.run_until_complete(test_get_data_us())

        # Test UK Units
        normal_config: Configuration = Configuration(
            auth_token=TOKEN,
            device_id=NORMAL_DEVICE_ID_UK,
            url=str(client.make_url(API_URL)),
            session=client.session,
        )
        interface: Interface = Interface(normal_config)

        async def test_get_data_uk():
            resp: PTDevicesResponse = await interface.get_data()
            assert resp.get("body") == GOOD_RESP_DICT_UK_IMP

        loop.run_until_complete(test_get_data_uk())
        loop.run_until_complete(client.close())


def test_resp_error_handling(test_web_app):
    """Test the interface with different response statuses."""
    with loop_context() as loop:
        # Setup the test server
        server = TestServer(test_web_app)
        client = TestClient(server, loop=loop)

        loop.run_until_complete(client.start_server())

        def make_interface(device_id: str) -> Interface:
            return Interface(
                Configuration(
                    auth_token=TOKEN,
                    device_id=device_id,
                    url=str(client.make_url(API_URL)),
                    session=client.session,
                )
            )

        async def test_errors():
            # UNAUTHORIZED (401) Handling
            interface: Interface = make_interface(UNAUTHORIZED_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesUnauthorizedError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # FORBIDDEN (403) Handling
            interface: Interface = make_interface(FORBIDDEN_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesForbiddenError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # FORBIDDEN (302) Handling
            interface: Interface = make_interface(REDIRECT_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesUnauthorizedError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # PTDevicesRequestError (403) Handling
            interface: Interface = make_interface(BAD_GATEWAY_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesRequestError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # Wrong content type Handling
            interface: Interface = make_interface(WRONG_CONTENT_TYPE_DEVICE_ID)

            with pytest.raises(PTDevicesRequestError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # FORBIDDEN (302) Handling
            interface: Interface = make_interface(EMPTY_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesRequestError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            with patch(
                "aiohttp.ClientSession.request",
                side_effect=TimeoutError,
            ):
                with pytest.raises(PTDevicesRequestError):
                    await interface.get_data()

        loop.run_until_complete(test_errors())
        loop.run_until_complete(client.close())


def test_real_server():
    """Test the interface with the real server."""
    with loop_context() as loop:
        # Setup the test server

        async def test_real_server():
            session: ClientSession = ClientSession(cookie_jar=CookieJar(unsafe=True))
            interface: Interface = Interface(
                Configuration(
                    auth_token=secret_TOKEN,
                    device_id=secret_DEVICE_ID,
                    url="https://api.ptdevices.com/token/v1",
                    session=session,
                )
            )
            resp: PTDevicesResponse = await interface.get_data()
            data: dict[str, dict[str, Any]] = resp.get("body", {})
            assert data
            assert "device_id" in list(data.values())[0]
            await session.close()

        loop.run_until_complete(test_real_server())


def test_real_server_multi():
    """Test the interface with the real server."""
    with loop_context() as loop:
        # Setup the test server

        async def test_real_server():
            session: ClientSession = ClientSession(cookie_jar=CookieJar(unsafe=True))
            interface: Interface = Interface(
                Configuration(
                    auth_token=secret_TOKEN,
                    device_id="*",
                    url="https://api.ptdevices.com/token/v1",
                    session=session,
                )
            )
            resp: PTDevicesResponse = await interface.get_data()
            data: dict[str, dict[str, Any]] = resp.get("body", {})
            assert data
            assert "device_id" in list(data.values())[0]
            await session.close()

        loop.run_until_complete(test_real_server())



"""Example for basic usage."""

from aiohttp import ClientSession

from aioptdevices import Configuration, Interface, PTDevicesResponse


async def main():

    web_session = ClientSession()

    interface = Interface(
        Configuration(
            auth_token="your API token",
            device_id="your device id or mac address",
            url="https://www.ptdevices.com/token/v1/device/",
            session=web_session,
        )
    )

    data: PTDevicesResponse = await interface.get_data()

    print(
        data.get("body")
    )  # Prints the device data, see https://support.paremtech.com/portal/en/kb/articles/api-options#Token_API



"""Common Test Helpers used by PyTest."""

from aiohttp import web
import pytest

from .mock_api import (
    API_URL,
    BAD_GATEWAY_ERROR_DEVICE_ID,
    EMPTY_ERROR_DEVICE_ID,
    FORBIDDEN_ERROR_DEVICE_ID,
    NORMAL_DEVICE_ID,
    NORMAL_DEVICE_ID_US,
    NORMAL_DEVICE_ID_UK,
    NORMAL_DEVICE_MAC,
    REDIRECT_ERROR_DEVICE_ID,
    UNAUTHORIZED_ERROR_DEVICE_ID,
    WRONG_CONTENT_TYPE_DEVICE_ID,
    bad_gateway_response,
    content_type_invalid_response,
    empty_response,
    forbidden_response,
    good_response,
    good_response_us_imperial_conversions,
    good_response_uk_imperial_conversions,
    redirect_response,
    unauthorized_response,
)


@pytest.fixture(name="test_web_app")
def test_web_app() -> web.Application:
    """Create a test server for testing."""
    app = web.Application()
    app.router.add_get(f"{API_URL}/device/{NORMAL_DEVICE_ID}", good_response)  # type: ignore  # noqa: PGH003
    app.router.add_get(f"{API_URL}/device/{NORMAL_DEVICE_MAC}", good_response)  # type: ignore  # noqa: PGH003
    app.router.add_get(
        f"{API_URL}/device/{NORMAL_DEVICE_ID_US}",
        good_response_us_imperial_conversions,
    )  # type: ignore  # noqa: PGH003
    app.router.add_get(
        f"{API_URL}/device/{NORMAL_DEVICE_ID_UK}",
        good_response_uk_imperial_conversions,
    )  # type: ignore  # noqa: PGH003

    app.router.add_get(
        f"{API_URL}/device/{EMPTY_ERROR_DEVICE_ID}",
        empty_response,
    )  # type: ignore  # noqa: PGH003

    app.router.add_get(
        f"{API_URL}/device/{UNAUTHORIZED_ERROR_DEVICE_ID}",
        unauthorized_response,
    )  # type: ignore  # noqa: PGH003
    app.router.add_get(
        f"{API_URL}/device/{FORBIDDEN_ERROR_DEVICE_ID}",
        forbidden_response,
    )  # type: ignore  # noqa: PGH003
    app.router.add_get(
        f"{API_URL}/device/{BAD_GATEWAY_ERROR_DEVICE_ID}",
        bad_gateway_response,
    )  # type: ignore  # noqa: PGH003
    app.router.add_get(
        f"{API_URL}/device{WRONG_CONTENT_TYPE_DEVICE_ID}",
        content_type_invalid_response,
    )  # type: ignore  # noqa: PGH003
    app.router.add_get(
        f"{API_URL}/device/{REDIRECT_ERROR_DEVICE_ID}",
        redirect_response,
    )  # type: ignore  # noqa: PGH003

    return app



"""Mock API for testing package."""

from typing import Any

from aiohttp import RequestInfo, web

# Configs
API_URL: str = "/token/v1"  # Token API URL
TOKEN: str = "6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b"  # fake token for testing

# Device IDs
NORMAL_DEVICE_ID: str = "200"  # Always returns OK (200) and a good body
NORMAL_DEVICE_MAC: str = "123456789ABC"
NORMAL_DEVICE_ID_US: str = "200US"
NORMAL_DEVICE_ID_UK: str = "200UK"
EMPTY_ERROR_DEVICE_ID: str = "0"  # Always returns an empty (200)
WRONG_CONTENT_TYPE_DEVICE_ID: str = "199"  # Same as NORMAL but wrong content type
UNAUTHORIZED_ERROR_DEVICE_ID: str = "401"  # Always Returns an UNAUTHORIZED (401)
REDIRECT_ERROR_DEVICE_ID: str = "302"  # Always Returns an FOUND (302)
FORBIDDEN_ERROR_DEVICE_ID: str = "403"  # Always Returns a FORBIDDEN (403)
BAD_GATEWAY_ERROR_DEVICE_ID: str = "502"  # Always Returns a BAD_GATEWAY (502)

# Responses
GOOD_RESP: str = '{"data":{"id":200,"device_id":"123456789ABC","share_id":"fFAa523e","created":"Jul 11th 2024, 7:44 PM","user_id":1234,"device_type":"level","title":"Test Response Level","version":"100","lat":null,"lng":null,"address":null,"status":"Working","status_number":2,"delivery_notes":null,"units":"Metric","reported":"Jul 15th, 1:25 PM","tx_reported":"Jul 15th, 8:17 AM","last_updated_on":null,"wifi_signal":"100%","tx_signal":"-46.00dBm","device_data":{"percent_level":50,"battery_voltage":6.6,"battery_status":"Good","battery_status_number":1,"volume_level":100,"inch_level":10,"enclosure_temperature":22.2},"device_setup":{"depth":10,"power_x":30,"power_y":3048,"power_z":24,"shape":"rectangle","diameter":null,"width":null,"length":null},"temperature_units":"C"}}'
GOOD_RESP_US_IMP: str = '{"data":{"id":200,"device_id":"123456789ABC","share_id":"fFAa523e","created":"Jul 11th 2024, 7:44 PM","user_id":1234,"device_type":"level","title":"Test Response Level","version":"100","lat":null,"lng":null,"address":null,"status":"Working","status_number":2,"delivery_notes":null,"units":"US Imperial","reported":"Jul 15th, 1:25 PM","tx_reported":"Jul 15th, 8:17 AM","last_updated_on":null,"wifi_signal":"100%","tx_signal":"-46.00dBm","device_data":{"percent_level":50,"battery_voltage":6.6,"battery_status":"Good","battery_status_number":1,"volume_level":100,"inch_level":10,"enclosure_temperature":22.2,"probe_temperature":70},"device_setup":{"depth":10,"power_x":30,"power_y":3048,"power_z":24,"shape":"rectangle","diameter":null,"width":null,"length":null},"temperature_units":"F"}}'
GOOD_RESP_UK_IMP: str = '{"data":{"id":200,"device_id":"123456789ABC","share_id":"fFAa523e","created":"Jul 11th 2024, 7:44 PM","user_id":1234,"device_type":"level","title":"Test Response Level","version":"100","lat":null,"lng":null,"address":null,"status":"Working","status_number":2,"delivery_notes":null,"units":"British Imperial","reported":"Jul 15th, 1:25 PM","tx_reported":"Jul 15th, 8:17 AM","last_updated_on":null,"wifi_signal":"100%","tx_signal":"-46.00dBm","device_data":{"percent_level":50,"battery_voltage":6.6,"battery_status":"Good","battery_status_number":1,"volume_level":100,"inch_level":10,"enclosure_temperature":22.2},"device_setup":{"depth":10,"power_x":30,"power_y":3048,"power_z":24,"shape":"rectangle","diameter":null,"width":null,"length":null},"temperature_units":"F"}}'

GOOD_RESP_DICT: dict[str, Any] = {
    "123456789ABC": {
        "id": 200,
        "device_id": "123456789ABC",
        "share_id": "fFAa523e",
        "created": "Jul 11th 2024, 7:44 PM",
        "user_id": 1234,
        "device_type": "level",
        "title": "Test Response Level",
        "version": 100,
        "lat": None,
        "lng": None,
        "address": None,
        "status": "working",
        "status_number": 2,
        "delivery_notes": None,
        "units": "Metric",
        "reported": "Jul 15th, 1:25 PM",
        "tx_reported": "Jul 15th, 8:17 AM",
        "last_updated_on": None,
        "wifi_signal": 100,
        "tx_signal": -46.00,
        "percent_level": 50,
        "battery_voltage": 6.6,
        "battery_status": "good",
        "battery_status_number": 1,
        "volume_level": 100,
        "depth_level": 0.254,
        "enclosure_temperature": 22.2,
        "depth": 10,
        "power_x": 30,
        "power_y": 3048,
        "power_z": 24,
        "shape": "rectangle",
        "diameter": None,
        "width": None,
        "length": None,
        "temperature_units": "C",
    }
}

GOOD_RESP_DICT_US_IMP: dict[str, Any] = {
    "123456789ABC": {
        "id": 200,
        "device_id": "123456789ABC",
        "share_id": "fFAa523e",
        "created": "Jul 11th 2024, 7:44 PM",
        "user_id": 1234,
        "device_type": "level",
        "title": "Test Response Level",
        "version": 100,
        "lat": None,
        "lng": None,
        "address": None,
        "status": "working",
        "status_number": 2,
        "delivery_notes": None,
        "units": "US Imperial",
        "reported": "Jul 15th, 1:25 PM",
        "tx_reported": "Jul 15th, 8:17 AM",
        "last_updated_on": None,
        "wifi_signal": 100,
        "tx_signal": -46.00,
        "percent_level": 50,
        "battery_voltage": 6.6,
        "battery_status": "good",
        "battery_status_number": 1,
        "volume_level": 378.541178,
        "depth_level": 0.254,
        "enclosure_temperature": 22.2,
        "probe_temperature": 21.111,
        "depth": 10,
        "power_x": 30,
        "power_y": 3048,
        "power_z": 24,
        "shape": "rectangle",
        "diameter": None,
        "width": None,
        "length": None,
        "temperature_units": "F",
    }
}

GOOD_RESP_DICT_UK_IMP: dict[str, Any] = {
    "123456789ABC": {
        "id": 200,
        "device_id": "123456789ABC",
        "share_id": "fFAa523e",
        "created": "Jul 11th 2024, 7:44 PM",
        "user_id": 1234,
        "device_type": "level",
        "title": "Test Response Level",
        "version": 100,
        "lat": None,
        "lng": None,
        "address": None,
        "status": "working",
        "status_number": 2,
        "delivery_notes": None,
        "units": "British Imperial",
        "reported": "Jul 15th, 1:25 PM",
        "tx_reported": "Jul 15th, 8:17 AM",
        "last_updated_on": None,
        "wifi_signal": 100,
        "tx_signal": -46.00,
        "percent_level": 50,
        "battery_voltage": 6.6,
        "battery_status": "good",
        "battery_status_number": 1,
        "volume_level": 454.609,
        "depth_level": 0.254,
        "enclosure_temperature": 22.2,
        "depth": 10,
        "power_x": 30,
        "power_y": 3048,
        "power_z": 24,
        "shape": "rectangle",
        "diameter": None,
        "width": None,
        "length": None,
        "temperature_units": "F",
    }
}


async def good_response(request: RequestInfo) -> web.Response:
    """Return an OK (200)."""
    assert request.url.query.get("api_token") == TOKEN  # Verify the token is matching

    return web.Response(status=200, content_type="application/json", body=GOOD_RESP)


async def good_response_us_imperial_conversions(request: RequestInfo) -> web.Response:
    """Return an OK (200)."""
    assert request.url.query.get("api_token") == TOKEN  # Verify the token is matching

    return web.Response(
        status=200, content_type="application/json", body=GOOD_RESP_US_IMP
    )


async def good_response_uk_imperial_conversions(request: RequestInfo) -> web.Response:
    """Return an OK (200)."""
    assert request.url.query.get("api_token") == TOKEN  # Verify the token is matching

    return web.Response(
        status=200, content_type="application/json", body=GOOD_RESP_UK_IMP
    )


async def empty_response(request: RequestInfo) -> web.Response:
    """Return an empty response (200)."""
    assert request.url.query.get("api_token") == TOKEN  # Verify the token is valid

    return web.Response(status=200, content_type="application/json", body="")


async def content_type_invalid_response(request: RequestInfo) -> web.Response:
    """Return an OK (200) with the wrong content type."""
    return web.Response(status=200, body=GOOD_RESP)


async def unauthorized_response(requsest: RequestInfo) -> web.Response:
    """Return an UNAUTHORIZED (401) to any request."""
    return web.Response(status=401)


async def forbidden_response(requsest: RequestInfo) -> web.Response:
    """Return an UNAUTHORIZED (403) to any request."""
    return web.Response(status=403)


async def bad_gateway_response(requsest: RequestInfo) -> web.Response:
    """Return an UNAUTHORIZED (502) to any request."""
    return web.Response(status=502)


async def not_found_response(request: RequestInfo) -> web.Response:
    """Return a PAGE NOT FOUND (404) to any request."""
    return web.Response(status=404)


async def redirect_response(request: RequestInfo) -> web.Response:
    """Return a PAGE NOT FOUND (302) to any request."""
    return web.Response(status=302)



"""CLI for PTDevices."""

import argparse
import asyncio
from asyncio.timeouts import timeout
import json
import logging

from aiohttp import ClientSession, CookieJar

import aioptdevices
from aioptdevices.configuration import Configuration
from aioptdevices.interface import Interface

LOGGER = logging.getLogger(__name__)


async def connect(
    deviceID: str,
    authToken: str,
    url: str,
    webSession: ClientSession,
) -> Interface | None:
    """Set up and Connect to PTDevices."""

    # Setup interface to PTDevices
    interface: Interface = Interface(
        Configuration(
            authToken,
            deviceID,
            url,
            webSession,
        )
    )
    try:
        async with timeout(10):
            data = await interface.get_data()

            formatted_body: str = json.dumps(data.get("body"), indent=2)
            LOGGER.info("Data: %s", formatted_body)
    except aioptdevices.PTDevicesRequestError as err:
        LOGGER.warning("failed to connect to PTDevices server: %s", err)

    except aioptdevices.PTDevicesUnauthorizedError as err:
        LOGGER.warning("Unable to read device data because of bad token: %s", err)

    except aioptdevices.PTDevicesForbiddenError as err:
        LOGGER.warning("Unable, device does not belong to the token holder: %s", err)
    else:
        return interface
    return None


async def main(
    deviceID: str,
    authToken: str,
    url: str,
) -> None:
    """Configure a connection and get the token api data from server."""

    # ----------  Connecting To PTDevices  -----------

    LOGGER.info(
        "\n%s\n", "  Connecting To PTDevices  ".center(48, "-")
    )  # Output a section title for connecting

    # Create a web session for use when connecting
    session: ClientSession = ClientSession(cookie_jar=CookieJar(unsafe=True))
    # session = ClientSession()

    # Setup connection to PTDevices
    ptdevicesInterface = await connect(deviceID, authToken, url, session)

    if not ptdevicesInterface:  # Failed connection to PTDevices
        LOGGER.error("Failed to connect to PTDevices")
        await session.close()
        return

    # -----------------  Connected  ------------------

    LOGGER.info(
        "\n%s\n", "  Connected  ".center(48, "-")
    )  # Output a section title when connected
    LOGGER.info("Successfully connected to %s", deviceID)

    # try:
    #     data = await ptdevicesInterface.get_data()
    #     LOGGER.info(f"Device {deviceID} data: \n{data}")
    # except aioptdevices.PTDevicesRequestError as err:
    #     LOGGER.warning(f"failed to connect to PTDevices server: {err}")

    # except aioptdevices.PTDevicesUnauthorizedError as err:
    #     LOGGER.warning(f"Unable, to read device data because of bad token: {err}")

    # except aioptdevices.PTDevicesForbiddenError as err:
    #     LOGGER.warning(f"Unable, device does not belong to the token holder: {err}")

    await session.close()
    return


def starter():
    """Parse CLI Arguments and fetch device data."""
    default_url = "https://api.ptdevices.com/token/v1"

    # Parse cli args
    parser = argparse.ArgumentParser()
    parser.add_argument("authToken", type=str)
    parser.add_argument("-I", "--id", type=str, default="*")
    parser.add_argument("-U", "--url", type=str, default=default_url)
    parser.add_argument("-D", "--debug", action="store_true")
    args = parser.parse_args()

    # Set the log level
    LOG_LEVEL = logging.INFO
    if args.debug:
        LOG_LEVEL = logging.DEBUG

    logging.basicConfig(format="%(message)s", level=LOG_LEVEL)

    # Output the settings
    # LOGGER.info(
    #     f"deviceID: {args.deviceID}\nToken: {args.authToken}\nurl: {args.url}\ndebug: {args.debug}\n"
    # )

    # --------------------  ARGS  --------------------

    LOGGER.info("\n%s\n", "  ARGS  ".center(48, "-"))  # Output a section title for args
    LOGGER.info("deviceID: %s", args.id)
    # LOGGER.info("Token: %s", args.authToken)
    LOGGER.info("url: %s", args.url)
    LOGGER.info("debug: %s", args.debug)

    # Run the program
    try:
        asyncio.run(
            main(
                deviceID=args.id,
                authToken=args.authToken,
                url=args.url,
            )
        )

    except KeyboardInterrupt:
        LOGGER.info("Keyboard interrupt")


if __name__ == "__main__":
    starter()
# ----------------------  ARGS  ------------------



"""Python Library for communicating with PTDevices."""

from enum import StrEnum
from http import HTTPStatus
import logging
from typing import Any, TypedDict, Required

import orjson
import asyncio
from string import ascii_letters

from aioptdevices.errors import (
    PTDevicesForbiddenError,
    PTDevicesRequestError,
    PTDevicesUnauthorizedError,
)

from .configuration import Configuration

LOGGER = logging.getLogger(__name__)

type PTDevicesResponseData = dict[str, dict[str, Any]]


class PTDevicesResponse(TypedDict, total=False):
    """Typed Response from PTDevices."""

    body: Required[dict[str, dict[str, Any]]]
    code: Required[int]


class PTDevicesBatteryState(StrEnum):
    """Store keys for hte different battery_status states."""

    UNKNOWN = "unknown"
    GOOD = "good"
    LOW = "low"


_battery_value_translations: dict[int, str] = {
    1: PTDevicesBatteryState.GOOD,
    2: PTDevicesBatteryState.LOW,
}


class PTDevicesStatusStates(StrEnum):
    """Store keys for the different device_status states."""

    UNKNOWN = "unknown"
    WORKING = "working"
    NOT_CONNECTED_YET = "not_connected_yet"
    NOT_CONNECTED = "not_connected"
    TRANSMITTER_NOT_REPORTING = "transmitter_not_reporting"
    PRESS_TRANSMITTER_CONNECT_BUTTON = "press_transmitter_connect_button"
    POWER_INTERNET_OUT_OR_RECEIVER_NOT_WORKING = (
        "power_internet_out_or_receiver_not_working"
    )


_status_value_translations: dict[int, str] = {
    0: PTDevicesStatusStates.NOT_CONNECTED_YET,
    1: PTDevicesStatusStates.NOT_CONNECTED_YET,
    2: PTDevicesStatusStates.WORKING,
    3: PTDevicesStatusStates.NOT_CONNECTED,
    4: PTDevicesStatusStates.TRANSMITTER_NOT_REPORTING,
    5: PTDevicesStatusStates.POWER_INTERNET_OUT_OR_RECEIVER_NOT_WORKING,
    6: PTDevicesStatusStates.PRESS_TRANSMITTER_CONNECT_BUTTON,
}

_convert_to_number_keys: dict[str, type] = {
    "version": int,
    "wifi_signal": int,
    "tx_signal": float,
    "percent_level": float,
    "battery_voltage": float,
    "volume_level": float,
    "inch_level": float,
    "probe_temperature": float,
}


def _format_data(
    input: list[dict[str, Any]],
) -> dict[str, dict[str, Any]]:
    """Format the returned data into a more helpful form. This includes putting id's as keys and flattening the dict."""
    output: dict[str, dict[str, Any]] = {}

    # Add id's as keys
    devices: dict[str, dict[str, Any]] = {
        device.get("device_id", ""): device for device in input
    }

    # Flatten the dict

    # Recurse through key value pairs and convert from nested to flat
    def recurse(sub_dict: dict[str, dict[str, Any]]) -> dict[str, Any]:
        results: dict[str, Any] = {}

        for key, value in sub_dict.items():
            if isinstance(value, dict):
                results = {**results, **recurse(value)}
            else:
                results[key] = value

        return results

    for device_id, device in devices.items():
        output[device_id] = recurse(device)

    # At this point the data has been transformed into a flat dict
    # {"id":{...info...},"id2":{...info...}}

    for device_id, device in output.items():
        # Translate the device status number
        if device["status_number"] in _status_value_translations.keys():
            output[device_id]["status"] = str(
                _status_value_translations[device["status_number"]]
            )
        else:
            output[device_id]["status"] = str(
                PTDevicesStatusStates.UNKNOWN
            )  # TODO tests dont cover this

        # Translate the battery status number
        if device.get("battery_status_number") is not None:
            if device["battery_status_number"] in _battery_value_translations.keys():
                output[device_id]["battery_status"] = str(
                    _battery_value_translations[device["battery_status_number"]]
                )
            else:
                output[device_id]["battery_status"] = str(
                    PTDevicesBatteryState.UNKNOWN
                )  # TODO tests dont cover this

        # Change all measurements to float | int | str
        for key in _convert_to_number_keys.keys():
            if device.get(key) is not None and isinstance(device.get(key), str):
                output[device_id][key] = _convert_to_number_keys[key](
                    device.get(key, "").strip(ascii_letters + "%"),
                )

        # Change the inch_level to depth_level and convert to metric (meters)
        if device.get("inch_level") is not None:
            output[device_id]["depth_level"] = round(
                0.0254 * float(device.get("inch_level", 0.0)), 6
            )
            output[device_id].pop("inch_level")

        if device.get("volume_level") is not None:
            if device.get("units", "") == "US Imperial":
                output[device_id]["volume_level"] = round(
                    3.785411784 * float(device.get("volume_level", 0.0)), 6
                )

            if device.get("units", "") == "British Imperial":
                output[device_id]["volume_level"] = round(
                    4.54609 * float(device.get("volume_level", 0.0)), 6
                )

        if device.get("probe_temperature") is not None:
            if device.get("temperature_units", "") == "F":
                output[device_id]["probe_temperature"] = round(
                    (float(device.get("probe_temperature", 0.0)) - 32) * 0.5555555556, 3
                )

    return output


class Interface:
    """Interface for PTDevices."""

    def __init__(self, config: Configuration) -> None:
        """Initilize object variables."""
        self.config = config

    async def get_data(self) -> PTDevicesResponse:
        """Fetch device data from PTDevices server and format it."""
        # Request url: https://api.ptdevices.com/token/v1/device/{deviceId}?api_token={given_token}
        # Where
        #   {deviceId} is the numeric internal device id,
        #       found in the url https://www.ptdevices.com/device/level/{deviceId}
        #   {given_token} is the access token you were given
        #
        # Request url: https://api.ptdevices.com/token/v1/devices?api_token={given_token}
        # Where
        #   {given_token} is the access token you were given

        # Construct the URL differently for multi device and single device requests
        if self.config.device_id == "*":
            url = f"{self.config.url}/devices?api_token={self.config.auth_token}"
            LOGGER.debug(
                "Sending request to %s for data from all devices",
                self.config.url,
            )
        else:
            url = f"{self.config.url}/device/{self.config.device_id}?api_token={self.config.auth_token}"
            LOGGER.debug(
                "Sending request to %s for data from device %s",
                self.config.url,
                self.config.device_id,
            )

        try:
            async with asyncio.timeout(10):
                async with self.config.session.request(
                    "get",
                    url,
                    allow_redirects=False,
                ) as results:
                    LOGGER.debug(
                        "%s Received from %s",
                        results.status,
                        self.config.url,
                    )

                    # Check return code
                    if results.status == HTTPStatus.UNAUTHORIZED:  # 401
                        raise PTDevicesUnauthorizedError(
                            f"Request to {url.split('?api_token')[0]} failed, the token provided is not valid"
                        )
                    if results.status == HTTPStatus.FOUND:  # 302
                        # Back end currently returns a 302 when request is not authorized
                        raise PTDevicesUnauthorizedError(
                            f"Request to {url.split('?api_token')[0]} failed, the token provided is not valid (302)"
                        )

                    if results.status == HTTPStatus.FORBIDDEN:  # 403
                        raise PTDevicesForbiddenError(
                            f"Request to {url.split('?api_token')[0]} failed, token invalid for device {self.config.device_id}"
                        )

                    if results.status != HTTPStatus.OK:  # anything but 200
                        raise PTDevicesRequestError(
                            f"Request to {url.split('?api_token')[0]} failed, got unexpected response from server ({results.status})"
                        )

                    # Check content type
                    if (
                        results.content_type != "application/json"
                        or results.content_length == 0
                    ):
                        raise PTDevicesRequestError(
                            f"Failed to get device data, returned content is invalid. Type: {results.content_type}, content Length: {results.content_length}, content: {results.content}"
                        )

                    raw_json = await results.read()

                    body = orjson.loads(raw_json)
                    formatted_data: dict[str, dict[str, Any]] = _format_data(
                        body["data"] if type(body["data"]) is list else [body["data"]]
                    )

                    # Store the new data to the response and return
                    return PTDevicesResponse(code=results.status, body=formatted_data)

        except TimeoutError:
            # If the request timed out, throw a request error
            raise PTDevicesRequestError(
                f"Request to {url.split('?api_token')[0]} failed, request timed out"
            )
